from mlmnemonist.factory import FACTORY
from testing.config.config_drug import get_cfg_defaults
from testing.unittest2 import decoy_recurring, decoy_prep, decoy_run

if __name__ == '__main__':
    description = "This is unittest3 runner"
    try:
        runner = FACTORY.retrieve_experiment_runner('0')
        runner.implement_run(decoy_run)
    except FileNotFoundError as e:
        runner = FACTORY.create_experiment_runner(verbose=4,
                                                  description="This is the runner for unittest 3",
                                                  cfg_base=get_cfg_defaults(),
                                                  cache_token='0')
        runner.preprocessing_pipeline.update_function(decoy_prep)
        runner.preprocess()
        runner.recurring_pipeline.update_function(decoy_recurring)
        runner.implement_run(decoy_run)

    print("--- Now running hyper-runner ---")

    try:
        hyper_runner = FACTORY.retrieve_hyper_experiment_runner('1')
    except FileNotFoundError as e:
        print("CREATING NEW HYPERRUNNER!")
        hyper_runner = FACTORY.create_hyper_experiment_runner(cfg_base=get_cfg_defaults(),
                                                              cfg_palette_path='conf-drug-palette.yaml',
                                                              experiment_runners=[runner],
                                                              verbose=1,
                                                              description='This is a hyperrunner for unittest 3',
                                                              cache_token='1'
                                                              )

    score_dict = hyper_runner.full_search()
    print(score_dict)



exception_list = (AttributeError, FileNotFoundError)

if __name__ == '__main__':
    n = int(input())

    try:
        if n == 0:
            raise AttributeError
        elif n == 1:
            raise FileNotFoundError
        elif n == 2:
            raise Exception
        else:
            print("HOWDY!")
    except exception_list as e:
        print("Exception caught!")


import sys

from config.config_drug import get_cfg_defaults
from mlmnemonist.validation_tools.cfg_utils import _expand_cfg, expand_cfg
from yacs.config import CfgNode as ConfigurationNode


def test1():
    cfg = ConfigurationNode()
    cfg.A = ConfigurationNode()
    cfg.B = ConfigurationNode()
    cfg.C = ConfigurationNode()
    cfg.A.A = ConfigurationNode()
    cfg.A.A.MLM_BRANCH_1 = 100
    cfg.A.A.MLM_BRANCH_2 = 200
    cfg.A.A.MLM_BRANCH_3 = 300
    cfg.A.B = ConfigurationNode()
    cfg.A.B.MLM_BRANCH_1 = 1000
    cfg.A.B.MLM_BRANCH_2 = 3000

    cfg.B.MLM_BRANCH_1 = ConfigurationNode()
    cfg.B.MLM_BRANCH_2 = ConfigurationNode()
    cfg.B.MLM_BRANCH_1.A = ConfigurationNode()
    cfg.B.MLM_BRANCH_1.A.MLM_BRANCH_1 = 111
    cfg.B.MLM_BRANCH_1.A.MLM_BRANCH_2 = 222
    cfg.B.MLM_BRANCH_2.T = ConfigurationNode()
    cfg.B.MLM_BRANCH_2.T.MLM_BRANCH_1 = 111
    cfg.B.MLM_BRANCH_2.T.MLM_BRANCH_2 = 222

    cfg.C.MLM_BRANCH_1 = 11
    cfg.C.MLM_BRANCH_2 = 13
    for cfg_i, code in _expand_cfg(cfg):
        print(f"-- {code} --")
        print(cfg_i)
        print("---")


def test2():
    sys.setrecursionlimit(10000)

    mp = expand_cfg(get_cfg_defaults(),
                    cfg_dir='conf-test-branches.yaml',
                    save_directory='all-confs')
    for key in mp.keys():
        print(key)
        print(mp[key])
        print("-------------------")


def test3():
    sys.setrecursionlimit(10000)

    mp = expand_cfg(get_cfg_defaults(),
                    cfg_dir='conf-drug-palette.yaml',
                    save_directory='config/drug-tuning')
    for key in mp.keys():
        print(key)
        print(mp[key])
        print("-------------------")


if __name__ == '__main__':
    test3()




import pandas as pd
from mlmnemonist import ExperimentRunner
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm
from IPython.display import clear_output
import matplotlib.pyplot as plt
import numpy as np

def load_raw_data(runner: ExperimentRunner):
    train_name = runner.cfg.DATASET.TRAIN_NAME
    test_name = runner.cfg.DATASET.TEST_NAME
    runner.train_df = pd.read_csv(runner.reveal_true_path(train_name))
    runner.test_df = pd.read_csv(runner.reveal_true_path(test_name))

def process_data(runner: ExperimentRunner):
    runner.train_X = runner.train_df.drop(columns='median_house_value').to_numpy()
    runner.train_Y = runner.train_df['median_house_value'].to_numpy()
    runner.test_X = runner.test_df.drop(columns='median_house_value').to_numpy()
    runner.test_Y = runner.test_df['median_house_value'].to_numpy()



class MyMLP(nn.Module):
    """
    A self explanatory MLP model with ReLU activations and 2 hidden layers
    """
    def __init__(self, input_features: int, hidden_layer_1: int, hidden_layer_2: int):
        super(MyMLP, self).__init__()
        self.l1 = nn.Linear(input_features, hidden_layer_1)
        self.l2 = nn.Linear(hidden_layer_1, hidden_layer_2)
        self.l3 = nn.Linear(hidden_layer_2, 1)

    def forward(self, x):
        x = self.l1(x)
        x = F.relu(x)
        x = self.l2(x)
        x = F.relu(x)
        x = self.l3(x)
        return x


def setup_device(runner: ExperimentRunner):
    # extract the device from config
    device = runner.cfg.SOLVER.DEVICE
    if device == 'cpu':
        runner.device = torch.device('cpu')
    elif device == 'cuda' and torch.cuda.is_available():
        runner.device = torch.device('cuda')
    else:
        raise NotImplementedError(f"device {device} is not implemented!")


def setup_model(runner: ExperimentRunner):
    # extract the method from config
    method = runner.cfg.SOLVER.METHOD
    # construct the model from registry
    cfg_h_params = runner.cfg.MODEL.HYPER_PARAMETERS
    my_model = MyMLP(input_features=cfg_h_params.IN_FEATURES,
                     hidden_layer_1=cfg_h_params.H1,
                     hidden_layer_2=cfg_h_params.H2)
    # set the function in the cache to save weights
    my_model = runner.CACHE.SET_M('mlp-key', my_model)
    my_model.to(runner.device)
    if runner.verbose > 0:
        print("Model state dict")
        print(my_model)


def setup_training(runner: ExperimentRunner):
    my_model = runner.CACHE.GET_M('mlp-key').to(runner.device)
    runner.criterion = nn.MSELoss()
    if runner.cfg.SOLVER.OPTIMIZER_TYPE == 'adam':
        runner.optim = torch.optim.Adam(my_model.parameters(), lr=runner.cfg.SOLVER.LR)
    else:
        raise NotImplementedError(f"Optimizer type not implemented"
                                  f"{runner.cfg.SOLVER.OPTIMIZER_TYPE}")




def my_custom_run(runner: ExperimentRunner, show_freq=50):
    # Get model from cache
    my_model = runner.CACHE.GET_M('mlp-key')

    # Get the epoch number and history from cache
    # if it is not cached from before set it to zero
    epoch_i = runner.CACHE.SET_IFN('epoch_i', 0)
    loss_history = runner.CACHE.SET_IFN('loss-history', {'train': [], 'test': []})
    for epoch_i in range(epoch_i, 400):

        # Shuffle all the indices first
        inds = np.arange(runner.train_X.shape[0])
        np.random.shuffle(inds)

        batch_size = runner.cfg.SOLVER.BATCH_SIZE

        loss_train = []
        my_model.train()

        for L in range(0, runner.train_X.shape[0], batch_size): 
            R = min(L + batch_size, runner.train_X.shape[0])
            # Find the range [L:R]
            X = torch.from_numpy(runner.train_X[inds[L:R],:]).float().to(runner.device)
            y = torch.from_numpy(runner.train_Y[inds[L:R]]).float().to(runner.device)
            loss = runner.criterion(my_model(X).squeeze(), y)
            # Find the loss between the predicted batch and y
            loss.backward()
            runner.optim.step()
            loss_train.append(loss.detach().cpu().item())

        # Find the average loss and add it to the trainign loss history
        mean_loss = sum(loss_train) / len(loss_train)
        loss_history['train'].append(mean_loss)

        # Now we will check the loss on the whole test dataset
        loss_test = []
        # Shuffle all the indices first
        inds = np.arange(runner.test_X.shape[0])
        np.random.shuffle(inds)
        my_model.eval()
        with torch.no_grad():
            for L in range(0, runner.test_X.shape[0], batch_size):
                R = min(L + batch_size, runner.test_X.shape[0])
                X = torch.from_numpy(runner.test_X[inds[L:R], :]).float().to(runner.device)
                y = torch.from_numpy(runner.test_Y[inds[L:R]]).float().to(runner.device)
                loss = runner.criterion(my_model(X).squeeze(), y)
                loss_test.append(loss.detach().cpu().item())
        # Find the average loss and add them to loss history
        mean_loss = sum(loss_test) / len(loss_test)
        loss_history['test'].append(mean_loss)


        # Display the losses
        if (epoch_i + 1) % show_freq == 0:
            clear_output()
            plt.plot(list(range(len(loss_history['train']))), loss_history['train'], label='loss-train')
            plt.plot(list(range(len(loss_history['test']))), loss_history['test'], label='loss-test')
            plt.legend()
            plt.show()

        # Caching and saving checkpoints
        runner.CACHE.SET('epoch_i', epoch_i)
        runner.CACHE.SET('loss-history', loss_history)
        runner.CACHE.SET_M('mlp-key', my_model)
        runner.CACHE.SAVE()

    return loss_history['test'][-1]


#!usr/bin/bash python

import sys
from google.colab import drive
import subprocess
import os
import shutil

PROJ_NAME = 'ML-Mnemonist'
GIT_DIR = 'HamidrezaKmK'

data_dir = None

if __name__ == '__main__':
  env = os.environ.copy()
  if not os.path.exists('/content/drive'):
    print("Mounting drive...")
    drive.mount('/content/drive')
    print("Mount complete!")

  while True:
    opt = input("What are you trying to do? [clone/pull] ")
    if opt == 'clone':
      addr = f"https://github.com/{GIT_DIR}/{PROJ_NAME}"
      print(f"Trying to connect to {addr}")
      token = input("Enter token: ")
      addr = addr.replace('[TOKEN]', token)
      res = subprocess.run(['git', 'clone', addr], env=env, capture_output=True)
      print(res.stdout.decode())
      print(res.stderr.decode())
      break
    elif opt == 'pull':
      path = os.path.join('/content', PROJ_NAME)
      os.chdir(path)
      res = subprocess.run(['git', 'pull'], env=env, capture_output=True)
      print(res.stdout.decode())
      print(res.stderr.decode())
      break
    elif opt == '':
      print("Nothing happened!")
      break
  
  if not os.path.exists(f'/content/{PROJ_NAME}'):
    raise RuntimeError("No project repository available!")

  if not os.path.exists(f'/content/{PROJ_NAME}/.env'):
    print("Dotenv non-existant!")
    while True:
      resp = input("Do you want to enter the file in the prompt or copy it?\n[copy/write] ")
      if resp == 'copy':
        dir = input("Enter the directory to copy: ")
        shutil.copyfile(dir, f'/content/{PROJ_NAME}/.env')
      elif resp == 'write':
        print("Enter the lines in format ENV_VARIABLE_NAME=VALUE")
        print("End with: ENDFILE")
        with open(f'/content/{PROJ_NAME}/.env', 'w') as f:
          while True:
            line = input()
            if line == 'ENDFILE':
              break
            f.write(f'{line}\n')
      else:
        continue
      break
        
  os.chdir('/content')






from mlmnemonist.experiment_runner import ExperimentRunner
from mlmnemonist.validation_tools.hyper_experiment_runner import HyperExperimentRunner
from mlmnemonist.factory import RunnerFactory
from mlmnemonist.factory import FACTORY
from mlmnemonist.runner_cache import MAX_CACHE_SIZE

__version__ = '0.1.0'


import json
import os
import warnings
from typing import List, Optional, Dict, Any, Tuple

from mlmnemonist.runner_cache import RunnerCache
from yacs.config import CfgNode as ConfigurationNode

from mlmnemonist import ExperimentRunner
from mlmnemonist.validation_tools import expand_cfg
from mlmnemonist.validation_tools.grid_search import _run_grid_search
from mlmnemonist.validation_tools.utils import get_all_codes


class HyperExperimentRunner:
    def __init__(self,
                 cfg_palette_dir: str,
                 cfg_base: ConfigurationNode,
                 hyper_experiment_path: str,
                 experiment_runners: List[ExperimentRunner],
                 checkpoint_dir: str,
                 cache_token: Optional[str] = None,
                 verbose: Optional[int] = 0,
                 secret_root: Optional[str] = None
                 ):
        """
        :param cfg_palette_dir: Path to the directory containing the configuration palettes
        :param cfg_base: Base configuration node that will be used to merge the configuration palettes
        :param hyper_experiment_path: Path to the directory where the hyper-experiment will store results
        :param experiment_runners: List of experiment runners that will be used to run the hyper-experiment
        :param checkpoint_dir: Path to the directory where the hyper-experiment will store cached data regarding itself
        :param cache_token: Token to be used to identify the hyper-experiment in the cache
        :param verbose: Verbosity level
        :param secret_root: The secret root that is being used per machine
        """
        self.cfg_base = cfg_base
        self._secret_root = secret_root
        self.hyper_experiment_path = hyper_experiment_path
        if not os.path.exists(self.hyper_experiment_path):
            raise FileNotFoundError(
                f"File {self.hyper_experiment_path} not found! Maybe you have deleted the experiment ...\n"
                f"Make sure to delete everything in checkpoint directory {checkpoint_dir} too before re-running")
        # Store all the configurations to run in a directory under hyper_experiment_path
        self.all_cfgs_dir = os.path.join(self.hyper_experiment_path, 'all-cfgs')
        if not os.path.exists(self.all_cfgs_dir):
            os.mkdir(self.all_cfgs_dir)
        expand_cfg(cfg_base, cfg_palette_dir, self.all_cfgs_dir)

        # raise exception if experiment_runners has length larger than 1
        if len(experiment_runners) > 1:
            raise NotImplementedError("Currently only one experiment runner is supported")
        self.runners = experiment_runners
        self.verbose = verbose
        self.checkpoint_dir = checkpoint_dir
        
        # Create a cache for the hyper-experiment
        self.CACHE = RunnerCache(self.checkpoint_dir, f'{cache_token}')

    def reveal_true_path(self, path: str) -> str:
        return os.path.join(self._secret_root, path)

    def preprocess(self, keep: bool = True) -> None:
        # run preprocess for all runners
        for runner in self.runners:
            runner.preprocess(keep=keep)

    def get_runner(self, runner_id: int = 0) -> ExperimentRunner:
        """
        Get an available experiment runner
        """
        return self.runners[runner_id]

    def run_code(self, code: str, with_preprocess: bool = False,
            exception_list: Tuple = (), *args, **kwargs):
        """
        Run a specific configuration on one of the runners that is available
        The runner might encounter an internal exception, in which case it will
        be caught from exception_list and ignored.

        This function updates the `score_dict` value available in the cache
        and returns the score of the configuration.
        """
        # Get the path to the configuration file
        all_codes = get_all_codes(self.all_cfgs_dir)
        cfg_path = all_codes[code]
        # Run the experiment
        runner = self.get_runner()
        runner.merge_cfg(cfg_path)
        # Setup the outputs
        current_experiment_output_dir = os.path.join(self.hyper_experiment_path, f'exp-{code}')
        if not os.path.exists(current_experiment_output_dir):
            os.mkdir(current_experiment_output_dir)
        runner.set_output_dir(current_experiment_output_dir)

        # Run the experiment
        if self.verbose > 0:
            print("---")
            print("This the configuration that will be used:")
            print(runner.cfg)
            print("---")
        runner.verbose = max(0, self.verbose - 1)
        try:
            if with_preprocess:
                runner.preprocess()
            score = runner.run(*args, **kwargs)
        except exception_list as e:
            warnings.warn(f"Exception caught {e}")
            score = None
        self.CACHE.LOAD()
        score_dict = self.CACHE.SET_IFN('score_dict', {})
        score_dict[code] = score
        self.CACHE.SET('score_dict', score_dict)
        self.CACHE.SAVE()
        runner.CACHE.RESET(prompt=False)
        return score


    def save_score_dict(self) -> None:
        """
        Save the score dictionary to the hyper-experiment directory
        """
        self.CACHE.LOAD()
        score_dict = self.CACHE.GET('score_dict')
        with open(os.path.join(self.hyper_experiment_path, 'score_dict.json'), 'w') as f:
            json.dump(score_dict, f)

    def full_search(self,
                    exception_list: Tuple = (),
                    with_preprocess=False,
                    *args, **kwargs) -> None:
        """
        :param exception_list: Some of the configurations might have inhenerent
        problems that will cause the experiment to crash. This list contains
        the codes of the configurations that should be skipped.

        :param with_preprocess: Whether to run the preprocess function before
        running the experiment. This is useful if the preprocess function
        generates some data that is needed for the experiment and we don't
        want to run the preprocess function again for each and every runner.

        :param args: Arguments to be passed to the experiment runner(s)
        :param kwargs: Keyword arguments to be passed to the experiment runner(s)
        """

        # Setup remaining codes
        self.CACHE.LOAD()
        iteration_i = self.CACHE.SET_IFN('iteration_i', 0)
        codes_dict = get_all_codes(self.all_cfgs_dir)
        all_codes = list(codes_dict.keys())
        self.CACHE.SAVE()

        # Iterate over all the codes and run the experiment while saving the end result in a json file
        while iteration_i < len(all_codes):
            current_code = all_codes[iteration_i]
            if self.verbose > 0:
                print(f"Iteration no. [{iteration_i + 1}/{len(all_codes)}]")
            self.run_code(current_code, with_preprocess=with_preprocess,exception_list=exception_list, *args, **kwargs)
            self.save_score_dict()
            iteration_i += 1
            self.CACHE.SET('iteration_i', iteration_i)
            self.CACHE.SAVE()

    def _grid_search(self, with_preprocess: bool = False, *args, **kwargs):
        # This version is unstable and probably buggy!
        return _run_grid_search(self.get_runner(),
                                self.CACHE,
                                self.verbose,
                                self.cfg_base,
                                self.all_cfgs_dir,
                                with_preprocess, *args, **kwargs)



from mlmnemonist.factory import FACTORY
from testing.config.config_drug import get_cfg_defaults
from testing.unittest2 import decoy_recurring, decoy_prep, decoy_run

if __name__ == '__main__':
    description = "This is unittest3 runner"
    try:
        runner = FACTORY.retrieve_experiment_runner('10')
    except FileNotFoundError as e:
        print("CREATING NEW!")
        runner = FACTORY.create_experiment_runner(verbose=4,
                                                  description="sth sth",
                                                  cfg_base=get_cfg_defaults(),
                                                  cache_token='10')
        runner.preprocessing_pipeline.update_function(decoy_prep)
        runner.recurring_pipeline.update_function(decoy_recurring)
        runner.implement_run(decoy_run)

    runner.preprocess()

    print("PREP DONE!")

    runner.run()


from typing import List, Optional, Union, Callable

from mlmnemonist.runner_cache import RunnerCache


class Pipeline:
    """
    This class contains a set of functions piped together.
    """

    def __init__(self, cache: Optional[RunnerCache] = None, name_in_cache: Optional[str] = None):
        self._all_functions: List[Callable] = []
        if cache is not None:
            self._CACHE = cache
            if name_in_cache is None:
                raise Exception("Cache defined but no name for pipeline!")
            self._name_in_cache = name_in_cache

    def __str__(self):
        return [x.__name__ for x in self._all_functions].__str__()

    def _find_index_in_pipeline(self, func: Callable) -> Optional[int]:
        """
        Find the index of function func in the pipeline
        """
        for i in range(len(self._all_functions)):
            f = self._all_functions[i]
            if f.__name__ == func.__name__:
                return i
        return None

    @property
    def function_count(self):
        """
        :return:
        The number of functions in the pipeline
        """
        return len(self._all_functions)

    def update_function(self, func: Callable) -> None:
        """
        This function updates the func if available and if not available
        then simply appends the function in the preprocessing pipeline
        """
        i = self._find_index_in_pipeline(func)
        if i is None:
            self.add_function(func)
        else:
            self._all_functions[i] = func
        self._CACHE.SET(self._name_in_cache, self._all_functions)
        self._CACHE.SAVE()

    def insert_function(self, func: Callable, index: int) -> None:
        """
        insert func at index 'index' of the pipeline 
        """
        if self._find_index_in_pipeline(func) is not None:
            raise Exception("Function is repeated!")
        self._all_functions.insert(index, func)
        self._CACHE.SET(self._name_in_cache, self._all_functions)
        self._CACHE.SAVE()

    def add_function(self, func: Callable) -> None:
        if self._find_index_in_pipeline(func) is not None:
            raise Exception("Function is repeated")
        self._all_functions.append(func)
        self._CACHE.SET(self._name_in_cache, self._all_functions)
        self._CACHE.SAVE()

    def remove_function(self, func: Callable) -> None:
        """
        Removes the function from pipeline if it is currently available in the pipeline
        """
        if self._find_index_in_pipeline(func) is not None:
            self._all_functions.remove(func)
        self._CACHE.SET(self._name_in_cache, self._all_functions)
        self._CACHE.SAVE()

    def clear_functions(self) -> None:
        """
        Clear the preprocessing pipeline
        """
        self._all_functions.clear()
        self._CACHE.SET(self._name_in_cache, self._all_functions)
        self._CACHE.SAVE()

    def run(self, keep: bool, verbose: int, runner, *args, **kwargs) -> None:
        """
        Runs the whole pipeline. If the keep flag is set to true, then the pipeline would remain the same
        after running. However, if it set to false, then all the things that are ran will be removed.
        """
        begin = True
        nxt_layer = None
        for i, f in enumerate(self._all_functions):
            if verbose > 0:
                print(f"[{i + 1}/{len(self._all_functions)}] Running {f.__name__}")
            if begin:
                nxt_layer = f(runner, *args, **kwargs)
                begin = False
            else:
                if nxt_layer is not None:
                    nxt_layer = f(runner, nxt_layer)
                else:
                    nxt_layer = f(runner)
        if not keep:
            self.clear_functions()
        return nxt_layer



# -*- coding: UTF-8 -*-

import io
import os.path
import random
import sys
import time
from typing import Optional, List, Any, Tuple

import httplib2
import openpyxl
from apiclient import discovery
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaIoBaseDownload, MediaIoBaseUpload

from drive import mimetypes
from drive.auth import authorize, get_credentials
from drive.exceptions import FileNotFoundException
from drive.files import File

# Retry transport and file IO errors.
RETRYABLE_ERRORS = (httplib2.HttpLib2Error, IOError)

# Number of times to retry failed downloads.
NUM_RETRIES = 5

# Number of bytes to send/receive in each request.
CHUNKSIZE = 2 * 1024 * 1024


def handle_progressless_iter(error, progressless_iters):
    if progressless_iters > NUM_RETRIES:
        print('Failed to make progress for too many consecutive iterations.')
        raise error

    sleeptime = random.random() * (2 ** progressless_iters)
    print('Caught exception (%s). Sleeping for %s seconds before retry #%d.'
          % (str(error), sleeptime, progressless_iters))
    time.sleep(sleeptime)


def print_with_carriage_return(s):
    """
    Internal utility to print a one-line string prefixed with a carriage return (``\\r``).
    :param s: string to print
    :return: None
    """
    sys.stdout.write('\r' + s)
    sys.stdout.flush()


def _make_querystring(clauses: List[Tuple[str, str, Any]], join="and"):
    """
    Make a "and" query string by combining all clauses. Each clause is a
    3-elements tuple of ``(field, operator, value)``. Refer to the
    following link for more information:
        https://developers.google.com/drive/v3/web/search-parameters
    :param clauses:
    :param join:
    :return:
    """
    parts = []
    for field, op, value in clauses:
        parts.append(_make_query_clause(field, op, value))

    return (" %s " % join).join(parts)


def _make_query_clause(field: str, op: str, value, negation=False) -> str:
    """

    :param field:
    :param op:
    :param value:
    :param negation:
    :return:
    """
    svalue = _serialize_query_value(value)
    if op == "in":
        p = "%s %s %s" % (svalue, op, field)
    else:
        p = "%s %s %s" % (field, op, svalue)
    if negation:
        p = "not %s" % p
    return p


def _serialize_query_value(value):
    """
    Serialize a query value.
    :param value:
    :return:
    """
    if isinstance(value, bool):
        return "true" if value else "false"

    return "'%s'" % str(value).replace("\\", "\\\\").replace("'", "\\'")


class Client:
    """
    Google Drive client
    """

    def __init__(self, credentials_path: Optional[str] = None):
        """

        :param credentials_path:
        """
        credentials = get_credentials(credentials_path)
        http = authorize(credentials)
        self.service = discovery.build('drive', 'v3', http=http)

    @property
    def _files(self):
        return self.service.files()

    @property
    def _permissions(self):
        return self.service.permissions()

    def create_folder(self, name: str, parent_id: str):
        """

        :param name:
        :param parent_id:
        :return:
        """
        file_metadata = {
            "name": name,
            "mimeType": mimetypes.GOOGLE_DRIVE_FOLDER,
        }
        if parent_id:
            file_metadata["parents"] = [parent_id]

        return self._execute_file_request(self._files.create(body=file_metadata))

    def get_or_create_folder(self, folder_name: str, parent_id: Optional[str] = None):
        """
        Get the ID for the folder with name folder_name, creating it if it doesn't exist.
        :param folder_name:
        :param parent_id:
        :return: drive.File
        """

        folder_list = self.list_files(name_equals=folder_name,
                                      mimetype=mimetypes.GOOGLE_DRIVE_FOLDER,
                                      parents_in=parent_id,
                                      n=1)
        if folder_list:
            if len(folder_list) == 1:
                return File(folder_list[0], client=self)

            raise NameError("Unable to find folder %s" % folder_name)

        return self.create_folder(folder_name, parent_id)

    def remove_file(self, file_id: str):
        """
        Remove a file by its id.

        :param file_id:
        :return:
        """
        return self._files.delete(fileId=file_id).execute()

    def get_file_metadata(self, file_id, raise_if_not_found=True, **kw):
        """

        :param file_id:
        :param raise_if_not_found:
        :param kw:
        :return:
        """
        try:
            return self._files.get(fileId=file_id, **kw).execute()
        except HttpError as e:
            if not raise_if_not_found:
                return None
            raise e

    def get_file(self, file_id: str, raise_if_not_found=True) -> Optional[File]:
        """
        Get a file by its id.
        :param file_id:
        :param raise_if_not_found: if ``True`` (default), raise an exception if the file doesn’t exist
        :return:
        """
        fm = self.get_file_metadata(file_id, raise_if_not_found)
        if fm:
            return File(fm, client=self)

    def get_file_by_name(self, name: str, parent_id: Optional[str] = None) -> Optional[File]:
        """
        Get a file by name.
        Note that, unlike ids, names are not guaranteed to be unique: you can have multiple files with the same name
        on Google Drive.

        :param name: Drive filename
        :param parent_id: optional parent id.
        :return:
        :raise: ``drive.exceptions.FileNotFoundException`` if the file doesn’t exist
        """
        kw = dict(name_equals=name, n=1)
        if parent_id:
            kw["parents_in"] = parent_id
        ls = self.list_files(**kw)
        if not ls:
            raise FileNotFoundException(name)

        return ls[0]

    def file_exists(self,
                    name: Optional[str] = None,
                    file_id: Optional[str] = None,
                    parent_id: Optional[str] = None) -> Optional[File]:
        """
        Check if a file exists and if so returns it.

        :param name:
        :param file_id:
        :param parent_id:
        :return:
        :raise: RuntimeError if both ``name`` and ``file_id`` are ``None``.
        """
        if not name and not file_id:
            raise RuntimeError("You must provide a name or file_id")

        if file_id:
            return self.get_file(file_id, raise_if_not_found=False)

        files = self.list_files(name_equals=name, parents_in=parent_id, n=1)
        if not files:
            return
        return files[0]

    def files_shared_with_me(self) -> List[File]:
        """
        Return a list of files (and 'directories') 'shared with me'.
        :return:
        """
        return self._execute_file_request(self._files.list(q="sharedWithMe=true"))

    def get_shared_file(self, name: str,
                        is_directory: Optional[bool] = None,
                        raise_if_not_found=True) -> Optional[File]:
        """
        Retrieve a shared file.
        If ``is_directory`` is a boolean, it’s used to filter files that are (or not) directories. By default the first
        matching file is returned without checking if it’s a directory or not.

        :param name:
        :param is_directory:
        :param raise_if_not_found:
        :return:
        """
        for shared in self.files_shared_with_me():
            if shared.name == name:
                if is_directory is False and shared.is_directory:
                    continue
                if is_directory and not shared.is_directory:
                    continue

                return shared

        if raise_if_not_found:
            raise FileNotFoundException(name)

    def get_shared_directory(self, name: str) -> Optional[File]:
        """
        Retrieve a shared directory. This is a shortcut for ``get_shared_file(name, is_directory=True)``.
        :param name:
        :return:
        """
        return self.get_shared_file(name, is_directory=True)

    def root(self) -> File:
        """
        Return the root directory. Note the alias ``"root"`` works as an alias file id for the root directory.
        :return:
        """
        return self.get_file("root")

    def list_files(self,
                   name_equals: Optional[str] = None,
                   name_contains: Optional[str] = None,
                   mimetype: Optional[str] = None,
                   parents_in=None,
                   n=100):
        """
        Outputs the names and IDs for up to N files.
        :param name_equals:
        :param name_contains:
        :param mimetype:
        :param parents_in:
        :param n:
        :return:
        """

        query_clauses = [("trashed", "=", False)]

        if name_equals:
            query_clauses.append(("name", "=", name_equals))
        if name_contains:
            query_clauses.append(("name", "contains", name_contains))
        if mimetype:
            query_clauses.append(("mimeType", "=", mimetype))
        if parents_in:
            query_clauses.append(("parents", "in", parents_in))

        q = _make_querystring(query_clauses)

        return self._execute_file_request(self._files.list(q=q, pageSize=n))

    def update_file(self, file_id: str,
                    remove_parents_ids=None,
                    add_parents_ids=None,
                    name: Optional[str] = None,
                    media=None):
        """

        :param file_id:
        :param remove_parents_ids:
        :param add_parents_ids:
        :param name:
        :param media:
        :return:
        """
        kw = dict(fileId=file_id)
        if remove_parents_ids:
            kw["removeParents"] = ",".join(remove_parents_ids)
        if add_parents_ids:
            kw["addParents"] = ",".join(add_parents_ids)
        if name:
            kw["body"] = {"name": name}

        if media:
            kw["media_body"] = media

        if len(kw) == 1:  # No modification, only fileId
            return

        return self._execute_file_request(self._files.update(**kw))

    def move_file_to_folder(self, file_id: str, folder_id: str):
        """

        :param file_id:
        :param folder_id:
        :return:
        """
        # Retrieve the existing parents to remove
        resp = self._files.get(fileId=file_id, fields='parents').execute()

        return self.update_file(file_id, add_parents_ids=[folder_id], remove_parents_ids=resp["parents"])

    def rename_file(self, file_id: str, name: str):
        """

        :param file_id:
        :param name:
        :return:
        """
        return self.update_file(file_id, name=name)

    def download(self, file_id: str, writer, mime_type: Optional[str] = None) -> None:
        """
        Download a file and write its content using the binary writer ``writer``. See also ``download_file``.

        Example:

            with open("my_file.ext", "wb") as f:
                client.download(file_id, f)

        :param file_id:
        :param writer: binary writer
        :param mime_type:
        :return:
        """
        kw = dict(fileId=file_id)
        fn = self._files.get_media

        if mime_type:
            kw["mimeType"] = mime_type
            fn = self._files.export_media

        downloader = MediaIoBaseDownload(writer, fn(**kw))
        # bypass the downloader; there appear to be a bug for large files
        writer.write(downloader._request.execute())

    def download_file(self, file_id: str, path: str, mime_type: Optional[str] = None) -> None:
        """
        Download a file and save it locally.
        :param file_id: file id
        :param path: local path where to save the file.
        :param mime_type: optional mime type
        :return:
        """
        with open(path, "wb") as f:
            self.download(file_id, f, mime_type=mime_type)

    def download_excel_workbook(self, file_id: str, read_only=False):
        """
        Download a Google Spreadsheet as an openpyxl workbook.
        :param file_id:
        :param read_only: set this to ``True`` if you don't plan to save or edit the workbook.
        :return: ``openpyxl.Workbook`` object.
        """
        buff = io.BytesIO()
        self.download(file_id, buff, mimetypes.XLSX)
        buff.seek(0)
        return openpyxl.load_workbook(buff, read_only=read_only)

    def upload(self, parent_id: str, name: str,
               reader,
               mime_type: Optional[str] = None,
               original_mime_type: Optional[str] = None,
               update_existing=False,
               resumable=False):
        """

        :param parent_id:
        :param name: remote filename
        :param reader: binary file reader
        :param mime_type:
        :param original_mime_type:
        :param update_existing:
        :param resumable:
        :return:
        """

        if isinstance(parent_id, File):
            parent_id = parent_id.id

        if not original_mime_type:
            import magic
            pos = reader.tell()
            buff = reader.read(1024)
            reader.seek(pos)
            original_mime_type = magic.from_buffer(buff, mime=True)

        media = MediaIoBaseUpload(reader, mimetype=original_mime_type,
                                  chunksize=CHUNKSIZE,
                                  resumable=resumable)

        if update_existing:
            f = self.file_exists(name=name, parent_id=parent_id)
            if f:
                return self.update_file(f.id, media=media)

        metadata = {
            'name': name,
            'parents': [parent_id],
        }

        if mime_type:
            metadata['mimeType'] = mime_type

        return self._execute_file_request(self._files.create(body=metadata,
                                                             media_body=media))

    def upload_file(self, parent_id: str, path: str,
                    name: Optional[str] = None,
                    mime_type: Optional[str] = None,
                    original_mime_type: Optional[str] = None,
                    update_existing=False):
        """

        :param parent_id:
        :param path: local path
        :param name: remote filename. If ``None``, use the local basename.
        :param mime_type:
        :param original_mime_type:
        :param update_existing:
        :return:
        """
        if isinstance(parent_id, File):
            parent_id = parent_id.id

        if name is None:
            name = os.path.basename(path)

        with open(path, "rb") as f:
            return self.upload(parent_id, name, f, mime_type,
                               original_mime_type,
                               update_existing=update_existing)

    def upload_excel_workbook(self,
                              parent: str,
                              name: str,
                              workbook: openpyxl.Workbook,
                              as_spreadsheet=True,
                              update_existing=False):
        """
        Upload an openpyxl (Excel) workbook and convert it to a Google Spreadsheet, unless ``as_spreadsheet`` is false.

        :param parent: parent id
        :param name: remote filename
        :param workbook: ``openpyxl.Workbook`` object
        :param as_spreadsheet: if ``True`` (default), convert the document to a Google Spreadsheet
        :param update_existing:
        :return:
        """
        buff = io.BytesIO()
        workbook.save(buff)
        buff.seek(0)

        target_mimetype = mimetypes.GOOGLE_SHEETS if as_spreadsheet else None

        return self.upload(parent, name, buff, target_mimetype,
                           mimetypes.XLSX, update_existing=update_existing)

    def grant_file_permissions(self, file_id: str, role: str, type_: str):
        """
        :param file_id:
        :param role:
        :param type_:
        :return:
        """
        return self._permissions.create(fileId=file_id, body={"role": role, "type": type_}).execute()

    def get_web_view_link(self, file_id: str) -> str:
        """
        :param file_id:
        :return:
        """
        resp = self._files.get(fileId=file_id, fields='webViewLink').execute()
        return resp['webViewLink']

    # Private API

    def _execute_file_request(self, req):
        """

        :param req:
        :return:
        """
        if not req.resumable:
            resp = req.execute()
            if "files" in resp:
                return [File(f, client=self) for f in resp["files"]]
            if "file" in resp:
                return File(resp["file"], client=self)
            return File(resp, client=self)
        else:
            progressless_iters = 0
            response = None
            progress = None
            while response is None:
                error = None
                try:
                    progress, response = req.next_chunk()
                    if progress:
                        print_with_carriage_return('Upload %d%%' %
                                                   (100 * progress.progress()))
                except HttpError as err:
                    error = err
                    if err.resp.status < 500:
                        raise
                except RETRYABLE_ERRORS as err:
                    error = err

                if error:
                    progressless_iters += 1
                    handle_progressless_iter(error, progressless_iters)
                else:
                    progressless_iters = 0

            if progress:
                print_with_carriage_return('Upload %d%%' %
                                           (100 * progress.progress()))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Challenges for reauthentication.
"""

import abc
import base64
import getpass
import sys

import six

from google.auth import _helpers
from google.auth import exceptions


REAUTH_ORIGIN = "https://accounts.google.com"


def get_user_password(text):
    """Get password from user.

    Override this function with a different logic if you are using this library
    outside a CLI.

    Args:
        text (str): message for the password prompt.

    Returns:
        str: password string.
    """
    return getpass.getpass(text)


@six.add_metaclass(abc.ABCMeta)
class ReauthChallenge(object):
    """Base class for reauth challenges."""

    @property
    @abc.abstractmethod
    def name(self):  # pragma: NO COVER
        """Returns the name of the challenge."""
        raise NotImplementedError("name property must be implemented")

    @property
    @abc.abstractmethod
    def is_locally_eligible(self):  # pragma: NO COVER
        """Returns true if a challenge is supported locally on this machine."""
        raise NotImplementedError("is_locally_eligible property must be implemented")

    @abc.abstractmethod
    def obtain_challenge_input(self, metadata):  # pragma: NO COVER
        """Performs logic required to obtain credentials and returns it.

        Args:
            metadata (Mapping): challenge metadata returned in the 'challenges' field in
                the initial reauth request. Includes the 'challengeType' field
                and other challenge-specific fields.

        Returns:
            response that will be send to the reauth service as the content of
            the 'proposalResponse' field in the request body. Usually a dict
            with the keys specific to the challenge. For example,
            ``{'credential': password}`` for password challenge.
        """
        raise NotImplementedError("obtain_challenge_input method must be implemented")


class PasswordChallenge(ReauthChallenge):
    """Challenge that asks for user's password."""

    @property
    def name(self):
        return "PASSWORD"

    @property
    def is_locally_eligible(self):
        return True

    @_helpers.copy_docstring(ReauthChallenge)
    def obtain_challenge_input(self, unused_metadata):
        passwd = get_user_password("Please enter your password:")
        if not passwd:
            passwd = " "  # avoid the server crashing in case of no password :D
        return {"credential": passwd}


class SecurityKeyChallenge(ReauthChallenge):
    """Challenge that asks for user's security key touch."""

    @property
    def name(self):
        return "SECURITY_KEY"

    @property
    def is_locally_eligible(self):
        return True

    @_helpers.copy_docstring(ReauthChallenge)
    def obtain_challenge_input(self, metadata):
        try:
            import pyu2f.convenience.authenticator
            import pyu2f.errors
            import pyu2f.model
        except ImportError:
            raise exceptions.ReauthFailError(
                "pyu2f dependency is required to use Security key reauth feature. "
                "It can be installed via `pip install pyu2f` or `pip install google-auth[reauth]`."
            )
        sk = metadata["securityKey"]
        challenges = sk["challenges"]
        app_id = sk["applicationId"]

        challenge_data = []
        for c in challenges:
            kh = c["keyHandle"].encode("ascii")
            key = pyu2f.model.RegisteredKey(bytearray(base64.urlsafe_b64decode(kh)))
            challenge = c["challenge"].encode("ascii")
            challenge = base64.urlsafe_b64decode(challenge)
            challenge_data.append({"key": key, "challenge": challenge})

        try:
            api = pyu2f.convenience.authenticator.CreateCompositeAuthenticator(
                REAUTH_ORIGIN
            )
            response = api.Authenticate(
                app_id, challenge_data, print_callback=sys.stderr.write
            )
            return {"securityKey": response}
        except pyu2f.errors.U2FError as e:
            if e.code == pyu2f.errors.U2FError.DEVICE_INELIGIBLE:
                sys.stderr.write("Ineligible security key.\n")
            elif e.code == pyu2f.errors.U2FError.TIMEOUT:
                sys.stderr.write("Timed out while waiting for security key touch.\n")
            else:
                raise e
        except pyu2f.errors.NoDeviceFoundError:
            sys.stderr.write("No security key found.\n")
        return None


AVAILABLE_CHALLENGES = {
    challenge.name: challenge
    for challenge in [SecurityKeyChallenge(), PasswordChallenge()]
}



from typing import List
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch_geometric.utils import to_dense_batch

from testing.modeling.registry import Registry
from yacs.config import CfgNode as ConfigurationNode

INTEGRATION_REGISTRY = Registry()


class DrugCellIntegrationModel(nn.Module):
    def __init__(self, drug_in_features: int, cell_in_features: int):
        super(DrugCellIntegrationModel, self).__init__()
        self.drug_in_features = drug_in_features
        self.cell_in_features = cell_in_features
        self.out_features = None


class SimpleConcatModel(DrugCellIntegrationModel):
    # TODO: add documentation
    def __init__(self, drug_in_features: int, cell_in_features: int, max_drug_count: int):
        super(SimpleConcatModel, self).__init__(drug_in_features, cell_in_features)
        self.k = max_drug_count
        self.out_features = self.k * drug_in_features + cell_in_features
        # self.batchnorm = nn.BatchNorm1d(self.out_features)

    def forward(self, all_drugs: List, cells):
        # all_drugs:
        #   [(d1 x drug_in_features), (d2 x drug_in_features), ..., (dB x drug_in_features)]
        # all di's should be less than or equal to max_drug_count
        # cells:
        #   (B x cell_in_features)
        B = len(all_drugs)
        all_padded = torch.stack([F.pad(d, (0, 0, 0, self.k - d.shape[0]), 'constant', 0) \
                                  for d in all_drugs])
        # all_padded.shape = (B * k * drug_in_features)
        all_padded = all_padded.reshape((B, -1))
        return torch.cat((all_padded, cells), dim=1)


class CellDrugAttentionModel(DrugCellIntegrationModel):
    def __init__(self, drug_in_features: int, cell_in_features: int, h: int, nheads: int):
        super(CellDrugAttentionModel, self).__init__(drug_in_features, cell_in_features)

        self.nheads = nheads
        self.attention_module = nn.MultiheadAttention(embed_dim=h * nheads,
                                                      num_heads=nheads,
                                                      kdim=drug_in_features,
                                                      vdim=drug_in_features,
                                                      batch_first=True)
        self.cell_query_fc = nn.Linear(cell_in_features, h * nheads)
        self.out_features = h * nheads + cell_in_features

    def forward(self, x_nodes, guide, cell):
        """
        x_nodes: all the nodes
        guide: A long tensor indicating each node belongs to which combination
        cells: B x cell
        """

        # Extract query, key, and value from drug and turn them into a batch_first sequence
        drug_key, mask = to_dense_batch(x_nodes, guide)

        # View cell lines as queries and drugs as key and value
        cell_query = self.cell_query_fc(cell.unsqueeze(1))
        cell_drug_integrated, _ = self.attention_module(query=cell_query,
                                                        key=drug_key,
                                                        value=drug_key,
                                                        need_weights=False,
                                                        key_padding_mask=torch.logical_not(mask)
                                                        )
        # B x M x h -> B x h
        cell_drug_integrated = cell_drug_integrated.squeeze()

        # Concatenate the max pooled feature vector from both sides
        integrated_all = torch.cat((cell_drug_integrated, cell), dim=1)
        return integrated_all


@INTEGRATION_REGISTRY.register('concatenation-method')
def build_concatenation_based_integration(cfg_hyper_parameters: ConfigurationNode, drug_in_features: int,
                                          cell_in_features: int):
    return SimpleConcatModel(
        drug_in_features=drug_in_features,
        cell_in_features=cell_in_features,
        max_drug_count=cfg_hyper_parameters.MAX_DRUG_COUNT
    )


@INTEGRATION_REGISTRY.register('multi-head-attention-method')
def build_attention_based_integration(cfg_hyper_parameters: ConfigurationNode, drug_in_features: int,
                                      cell_in_features: int):
    return CellDrugAttentionModel(
        drug_in_features=drug_in_features,
        cell_in_features=cell_in_features,
        nheads=cfg_hyper_parameters.N_HEADS,
        h=cfg_hyper_parameters.H
    )



import torch
import torch.nn as nn
import torch.nn.functional as F

from testing.modeling.registry import Registry
from yacs.config import CfgNode as ConfigurationNode

CELL_EMBEDDER_REGISTRY = Registry()


class CellEmbedder(nn.Module):
    def __init__(self, dropout: float = 0.2,
                 out_features: int = 100):
        super(CellEmbedder, self).__init__()
        self.dropout_rate = dropout
        self.out_features = out_features


class MLPCellEmbedder(CellEmbedder):
    def __init__(self, dropout: float,
                 out_features: int,
                 in_features: int,
                 h1: int = 2048,
                 h2: int = 512):
        super(MLPCellEmbedder, self).__init__(dropout, out_features)
        self.expression_layer1 = nn.Linear(in_features, h1, bias=False)
        self.batchnorm1 = nn.BatchNorm1d(h1)
        self.expression_layer2 = nn.Linear(h1, h2, bias=False)
        self.batchnorm2 = nn.BatchNorm1d(h2)
        self.expression_layer3 = nn.Linear(h2, out_features)

    def forward(self, expression_encodings):
        x = self.expression_layer1(expression_encodings)
        x = F.relu(x)
        x = self.batchnorm1(x)
        x = F.dropout(x, p=self.dropout_rate, training=self.training)
        x = self.expression_layer2(x)
        x = F.relu(x)
        x = self.batchnorm2(x)
        x = F.dropout(x, p=self.dropout_rate, training=self.training)
        x = self.expression_layer3(x)
        x = F.relu(x)
        return x


@CELL_EMBEDDER_REGISTRY.register('mlp-cell-embedder')
def build_simple_cell_embedder_with_expression(cfg_hyper_parameters: ConfigurationNode):
    return MLPCellEmbedder(dropout=cfg_hyper_parameters.DROPOUT,
                           out_features=cfg_hyper_parameters.OUT_FEATURES,
                           in_features=cfg_hyper_parameters.IN_FEATURES,
                           h1=cfg_hyper_parameters.H1,
                           h2=cfg_hyper_parameters.H2)



import os
from typing import Dict


def get_all_codes(parent_directory: str) -> Dict[str, str]:
    """
    This function is called on a parent directory that contains all the
    configuration files. It returns a dictionary of all the codes and their
    corresponding configuration file paths.

    For example:
    > parent_directory
        0-0-MLM-conf-0.yaml
        1-0-MLM-conf-0.yaml
        0-1-MLM-conf-0.yaml
    
    Furthermore, the function returns a dictionary of the form:
    {
        "0-0": "0-0-MLM-conf-0.yaml",
        "1-0": "1-0-MLM-conf-0.yaml",
        "0-1": "0-1-MLM-conf-0.yaml"
    }
    """
    all_codes = {}
    for f in os.listdir(parent_directory):
        if f.split('.')[-1] == 'yaml':
            all_codes[f.split('-MLM-')[0]] = os.path.join(parent_directory, f)

    return all_codes



"""
magic is a wrapper around the libmagic file identification library.

See README for more information.

Usage:

>>> import magic
>>> magic.from_file("testdata/test.pdf")
'PDF document, version 1.2'
>>> magic.from_file("testdata/test.pdf", mime=True)
'application/pdf'
>>> magic.from_buffer(open("testdata/test.pdf").read(1024))
'PDF document, version 1.2'
>>>

"""

import sys
import glob
import ctypes
import ctypes.util
import threading
import logging

from ctypes import c_char_p, c_int, c_size_t, c_void_p, byref, POINTER

# avoid shadowing the real open with the version from compat.py
_real_open = open


class MagicException(Exception):
    def __init__(self, message):
        super(Exception, self).__init__(message)
        self.message = message


class Magic:
    """
    Magic is a wrapper around the libmagic C library.
    """

    def __init__(self, mime=False, magic_file=None, mime_encoding=False,
                 keep_going=False, uncompress=False, raw=False, extension=False):
        """
        Create a new libmagic wrapper.

        mime - if True, mimetypes are returned instead of textual descriptions
        mime_encoding - if True, codec is returned
        magic_file - use a mime database other than the system default
        keep_going - don't stop at the first match, keep going
        uncompress - Try to look inside compressed files.
        raw - Do not try to decode "non-printable" chars.
        extension - Print a slash-separated list of valid extensions for the file type found.
        """
        self.flags = MAGIC_NONE
        if mime:
            self.flags |= MAGIC_MIME_TYPE
        if mime_encoding:
            self.flags |= MAGIC_MIME_ENCODING
        if keep_going:
            self.flags |= MAGIC_CONTINUE
        if uncompress:
            self.flags |= MAGIC_COMPRESS
        if raw:
            self.flags |= MAGIC_RAW
        if extension:
            self.flags |= MAGIC_EXTENSION

        self.cookie = magic_open(self.flags)
        self.lock = threading.Lock()

        magic_load(self.cookie, magic_file)

        # MAGIC_EXTENSION was added in 523 or 524, so bail if
        # it doesn't appear to be available
        if extension and (not _has_version or version() < 524):
            raise NotImplementedError('MAGIC_EXTENSION is not supported in this version of libmagic')

        # For https://github.com/ahupp/python-magic/issues/190
        # libmagic has fixed internal limits that some files exceed, causing
        # an error.  We can avoid this (at least for the sample file given)
        # by bumping the limit up.  It's not clear if this is a general solution
        # or whether other internal limits should be increased, but given
        # the lack of other reports I'll assume this is rare.
        if _has_param:
            try:
                self.setparam(MAGIC_PARAM_NAME_MAX, 64)
            except MagicException as e:
                # some versions of libmagic fail this call,
                # so rather than fail hard just use default behavior
                pass

    def from_buffer(self, buf):
        """
        Identify the contents of `buf`
        """
        with self.lock:
            try:
                # if we're on python3, convert buf to bytes
                # otherwise this string is passed as wchar*
                # which is not what libmagic expects
                # NEXTBREAK: only take bytes
                if type(buf) == str and str != bytes:
                    buf = buf.encode('utf-8', errors='replace')
                return maybe_decode(magic_buffer(self.cookie, buf))
            except MagicException as e:
                return self._handle509Bug(e)

    def from_file(self, filename):
        # raise FileNotFoundException or IOError if the file does not exist
        with _real_open(filename):
            pass

        with self.lock:
            try:
                return maybe_decode(magic_file(self.cookie, filename))
            except MagicException as e:
                return self._handle509Bug(e)

    def from_descriptor(self, fd):
        with self.lock:
            try:
                return maybe_decode(magic_descriptor(self.cookie, fd))
            except MagicException as e:
                return self._handle509Bug(e)

    def _handle509Bug(self, e):
        # libmagic 5.09 has a bug where it might fail to identify the
        # mimetype of a file and returns null from magic_file (and
        # likely _buffer), but also does not return an error message.
        if e.message is None and (self.flags & MAGIC_MIME_TYPE):
            return "application/octet-stream"
        else:
            raise e

    def setparam(self, param, val):
        return magic_setparam(self.cookie, param, val)

    def getparam(self, param):
        return magic_getparam(self.cookie, param)

    def __del__(self):
        # no _thread_check here because there can be no other
        # references to this object at this point.

        # during shutdown magic_close may have been cleared already so
        # make sure it exists before using it.

        # the self.cookie check should be unnecessary and was an
        # incorrect fix for a threading problem, however I'm leaving
        # it in because it's harmless and I'm slightly afraid to
        # remove it.
        if hasattr(self, 'cookie') and self.cookie and magic_close:
            magic_close(self.cookie)
            self.cookie = None


_instances = {}


def _get_magic_type(mime):
    i = _instances.get(mime)
    if i is None:
        i = _instances[mime] = Magic(mime=mime)
    return i


def from_file(filename, mime=False):
    """"
    Accepts a filename and returns the detected filetype.  Return
    value is the mimetype if mime=True, otherwise a human readable
    name.

    >>> magic.from_file("testdata/test.pdf", mime=True)
    'application/pdf'
    """
    m = _get_magic_type(mime)
    return m.from_file(filename)


def from_buffer(buffer, mime=False):
    """
    Accepts a binary string and returns the detected filetype.  Return
    value is the mimetype if mime=True, otherwise a human readable
    name.

    >>> magic.from_buffer(open("testdata/test.pdf").read(1024))
    'PDF document, version 1.2'
    """
    m = _get_magic_type(mime)
    return m.from_buffer(buffer)


def from_descriptor(fd, mime=False):
    """
    Accepts a file descriptor and returns the detected filetype.  Return
    value is the mimetype if mime=True, otherwise a human readable
    name.

    >>> f = open("testdata/test.pdf")
    >>> magic.from_descriptor(f.fileno())
    'PDF document, version 1.2'
    """
    m = _get_magic_type(mime)
    return m.from_descriptor(fd)

from . import loader
libmagic = loader.load_lib()

magic_t = ctypes.c_void_p


def errorcheck_null(result, func, args):
    if result is None:
        err = magic_error(args[0])
        raise MagicException(err)
    else:
        return result


def errorcheck_negative_one(result, func, args):
    if result == -1:
        err = magic_error(args[0])
        raise MagicException(err)
    else:
        return result


# return str on python3.  Don't want to unconditionally
# decode because that results in unicode on python2
def maybe_decode(s):
    # NEXTBREAK: remove
    if str == bytes:
        return s
    else:
        # backslashreplace here because sometimes libmagic will return metadata in the charset
        # of the file, which is unknown to us (e.g the title of a Word doc)
        return s.decode('utf-8', 'backslashreplace')


try:
    from os import PathLike
    def unpath(filename):
        if isinstance(filename, PathLike):
            return filename.__fspath__()
        else:
            return filename
except ImportError:
    def unpath(filename):
        return filename

def coerce_filename(filename):
    if filename is None:
        return None

    filename = unpath(filename)

    # ctypes will implicitly convert unicode strings to bytes with
    # .encode('ascii').  If you use the filesystem encoding
    # then you'll get inconsistent behavior (crashes) depending on the user's
    # LANG environment variable
    # NEXTBREAK: remove
    is_unicode = (sys.version_info[0] <= 2 and
                 isinstance(filename, unicode)) or \
                 (sys.version_info[0] >= 3 and
                  isinstance(filename, str))
    if is_unicode:
        return filename.encode('utf-8', 'surrogateescape')
    else:
        return filename


magic_open = libmagic.magic_open
magic_open.restype = magic_t
magic_open.argtypes = [c_int]

magic_close = libmagic.magic_close
magic_close.restype = None
magic_close.argtypes = [magic_t]

magic_error = libmagic.magic_error
magic_error.restype = c_char_p
magic_error.argtypes = [magic_t]

magic_errno = libmagic.magic_errno
magic_errno.restype = c_int
magic_errno.argtypes = [magic_t]

_magic_file = libmagic.magic_file
_magic_file.restype = c_char_p
_magic_file.argtypes = [magic_t, c_char_p]
_magic_file.errcheck = errorcheck_null


def magic_file(cookie, filename):
    return _magic_file(cookie, coerce_filename(filename))


_magic_buffer = libmagic.magic_buffer
_magic_buffer.restype = c_char_p
_magic_buffer.argtypes = [magic_t, c_void_p, c_size_t]
_magic_buffer.errcheck = errorcheck_null


def magic_buffer(cookie, buf):
    return _magic_buffer(cookie, buf, len(buf))


magic_descriptor = libmagic.magic_descriptor
magic_descriptor.restype = c_char_p
magic_descriptor.argtypes = [magic_t, c_int]
magic_descriptor.errcheck = errorcheck_null

_magic_descriptor = libmagic.magic_descriptor
_magic_descriptor.restype = c_char_p
_magic_descriptor.argtypes = [magic_t, c_int]
_magic_descriptor.errcheck = errorcheck_null


def magic_descriptor(cookie, fd):
    return _magic_descriptor(cookie, fd)


_magic_load = libmagic.magic_load
_magic_load.restype = c_int
_magic_load.argtypes = [magic_t, c_char_p]
_magic_load.errcheck = errorcheck_negative_one


def magic_load(cookie, filename):
    return _magic_load(cookie, coerce_filename(filename))


magic_setflags = libmagic.magic_setflags
magic_setflags.restype = c_int
magic_setflags.argtypes = [magic_t, c_int]

magic_check = libmagic.magic_check
magic_check.restype = c_int
magic_check.argtypes = [magic_t, c_char_p]

magic_compile = libmagic.magic_compile
magic_compile.restype = c_int
magic_compile.argtypes = [magic_t, c_char_p]

_has_param = False
if hasattr(libmagic, 'magic_setparam') and hasattr(libmagic, 'magic_getparam'):
    _has_param = True
    _magic_setparam = libmagic.magic_setparam
    _magic_setparam.restype = c_int
    _magic_setparam.argtypes = [magic_t, c_int, POINTER(c_size_t)]
    _magic_setparam.errcheck = errorcheck_negative_one

    _magic_getparam = libmagic.magic_getparam
    _magic_getparam.restype = c_int
    _magic_getparam.argtypes = [magic_t, c_int, POINTER(c_size_t)]
    _magic_getparam.errcheck = errorcheck_negative_one


def magic_setparam(cookie, param, val):
    if not _has_param:
        raise NotImplementedError("magic_setparam not implemented")
    v = c_size_t(val)
    return _magic_setparam(cookie, param, byref(v))


def magic_getparam(cookie, param):
    if not _has_param:
        raise NotImplementedError("magic_getparam not implemented")
    val = c_size_t()
    _magic_getparam(cookie, param, byref(val))
    return val.value


_has_version = False
if hasattr(libmagic, "magic_version"):
    _has_version = True
    magic_version = libmagic.magic_version
    magic_version.restype = c_int
    magic_version.argtypes = []


def version():
    if not _has_version:
        raise NotImplementedError("magic_version not implemented")
    return magic_version()


MAGIC_NONE = 0x000000  # No flags
MAGIC_DEBUG = 0x000001  # Turn on debugging
MAGIC_SYMLINK = 0x000002  # Follow symlinks
MAGIC_COMPRESS = 0x000004  # Check inside compressed files
MAGIC_DEVICES = 0x000008  # Look at the contents of devices
MAGIC_MIME_TYPE = 0x000010  # Return a mime string
MAGIC_MIME_ENCODING = 0x000400  # Return the MIME encoding
# TODO:  should be
# MAGIC_MIME = MAGIC_MIME_TYPE | MAGIC_MIME_ENCODING
MAGIC_MIME = 0x000010  # Return a mime string
MAGIC_EXTENSION = 0x1000000  # Return a /-separated list of extensions

MAGIC_CONTINUE = 0x000020  # Return all matches
MAGIC_CHECK = 0x000040  # Print warnings to stderr
MAGIC_PRESERVE_ATIME = 0x000080  # Restore access time on exit
MAGIC_RAW = 0x000100  # Don't translate unprintable chars
MAGIC_ERROR = 0x000200  # Handle ENOENT etc as real errors

MAGIC_NO_CHECK_COMPRESS = 0x001000  # Don't check for compressed files
MAGIC_NO_CHECK_TAR = 0x002000  # Don't check for tar files
MAGIC_NO_CHECK_SOFT = 0x004000  # Don't check magic entries
MAGIC_NO_CHECK_APPTYPE = 0x008000  # Don't check application type
MAGIC_NO_CHECK_ELF = 0x010000  # Don't check for elf details
MAGIC_NO_CHECK_ASCII = 0x020000  # Don't check for ascii files
MAGIC_NO_CHECK_TROFF = 0x040000  # Don't check ascii/troff
MAGIC_NO_CHECK_FORTRAN = 0x080000  # Don't check ascii/fortran
MAGIC_NO_CHECK_TOKENS = 0x100000  # Don't check ascii/tokens

MAGIC_PARAM_INDIR_MAX = 0  # Recursion limit for indirect magic
MAGIC_PARAM_NAME_MAX = 1  # Use count limit for name/use magic
MAGIC_PARAM_ELF_PHNUM_MAX = 2  # Max ELF notes processed
MAGIC_PARAM_ELF_SHNUM_MAX = 3  # Max ELF program sections processed
MAGIC_PARAM_ELF_NOTES_MAX = 4  # # Max ELF sections processed
MAGIC_PARAM_REGEX_MAX = 5  # Length limit for regex searches
MAGIC_PARAM_BYTES_MAX = 6  # Max number of bytes to read from file


# This package name conflicts with the one provided by upstream
# libmagic.  This is a common source of confusion for users.  To
# resolve, We ship a copy of that module, and expose it's functions
# wrapped in deprecation warnings.
def _add_compat(to_module):
    import warnings, re
    from magic import compat

    def deprecation_wrapper(fn):
        def _(*args, **kwargs):
            warnings.warn(
                "Using compatibility mode with libmagic's python binding. "
                "See https://github.com/ahupp/python-magic/blob/master/COMPAT.md for details.",
                PendingDeprecationWarning)

            return fn(*args, **kwargs)

        return _

    fn = ['detect_from_filename',
          'detect_from_content',
          'detect_from_fobj',
          'open']
    for fname in fn:
        to_module[fname] = deprecation_wrapper(compat.__dict__[fname])

    # copy constants over, ensuring there's no conflicts
    is_const_re = re.compile("^[A-Z_]+$")
    allowed_inconsistent = set(['MAGIC_MIME'])
    for name, value in compat.__dict__.items():
        if is_const_re.match(name):
            if name in to_module:
                if name in allowed_inconsistent:
                    continue
                if to_module[name] != value:
                    raise Exception("inconsistent value for " + name)
                else:
                    continue
            else:
                to_module[name] = value


_add_compat(globals())
